@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: dns.c,v 1.59 1998/09/29 02:09:28 motonori Exp $";
+static char *_id_ = "$Id: dns.c,v 1.70 1999/10/16 07:55:15 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -48,15 +48,23 @@ static char *_id_ = "$Id: dns.c,v 1.59 1998/09/29 02:09:28 motonori Exp $";
 #endif
 
 #ifdef ultrix
-#define NSADDR_LIST(x)	ns_list[x].addr	/* ultrix implementation is strange */
+#define NSADDR_LIST(x)	_res.ns_list[x].addr	/* ultrix implementation is strange */
 #endif
 
 #ifndef NSADDR_LIST
-#ifdef INET6
-#define NSADDR_LIST(x)	nsaddr_list_un[x]	/* for standerd BIND */
-#else
-#define NSADDR_LIST(x)	nsaddr_list[x]		/* for standerd BIND */
-#endif
+# if defined(INET6) && !defined(RESOLVER_V4)
+#  ifdef RESOLVER_KAME
+#   define NSADDR_LIST(x)	_res_ext.nsaddr_list[x]	/* for standerd BIND */
+#  else
+#   ifdef RESOLVER_OLDKAME
+#    define NSADDR_LIST(x)	_res.nsaddr_list_un[x]	/* for standerd BIND */
+#   else
+#    error unknown resolver type
+#   endif
+#  endif
+# else
+#  define NSADDR_LIST(x)	_res.nsaddr_list[x]		/* for standerd BIND */
+# endif
 #endif
 
 #define CNAMELOOPMAX	10
@@ -80,15 +88,18 @@ struct domain *dl;
 	long stime, etime;
 	int n;
 	static int query_domain(), process_query();
+#ifdef INET6
+	SockAddr dsin;	/* dummy */
+#endif
 
 	for (n = 0; n < _res.nscount; n++)
 	{
 		struct sockaddr_in *srv;
 
 #if INET6
-		srv = &_res.NSADDR_LIST(n).sin;
+		srv = (struct sockaddr_in *)&NSADDR_LIST(n);
 #else
-		srv = &_res.NSADDR_LIST(n);
+		srv = &NSADDR_LIST(n);
 #endif
 		if (srv->sin_family == AF_INET
 		 && srv->sin_addr.s_addr == INADDR_ANY)
@@ -106,42 +117,96 @@ struct domain *dl;
 	else
 		setproctitle("pre-getting MX", NULL);
 
-	if (*fallbackmx)
+	if (cnf.fallbackmx != NULL)
 	{
 		if (cnf.debug & DEBUG_DNS)
-		log(LOG_DEBUG, "pregetting A of fallbackmx: %s", fallbackmx);
+		log(LOG_DEBUG, "pregetting A of fallbackmx: %s",
+			cnf.fallbackmx);
 #ifdef INET6
 		if (cnf.inetdom & SMTP_V4)
 		{
 #endif
-			if (query_domain(fallbackmx, T_A) < 0)
+			if (query_domain(cnf.fallbackmx, T_A) < 0)
 				return -1;
 			n++;
 #ifdef INET6
 		}
 		if (cnf.inetdom & SMTP_V6)
 		{
-			if (query_domain(fallbackmx, T_AAAA) < 0)
+			if (query_domain(cnf.fallbackmx, T_AAAA) < 0)
 				return -1;
 			n++;
 		}
 #endif
 	}
 
+	if (cnf.sourceIP != NULL)
+	{
+		char *p, *q;
+
+		p = cnf.sourceIP;
+		while (p != NULL)
+		{
+			q = strchr(p, ',');
+			if (q != NULL)
+				*q = '\0';
+
+#ifdef INET6
+			if (inet_addr(p) == -1
+			 && inet_pton(INET6, p, &dsin.in6.sin6_addr) <= 0)
+#else
+			if (inet_addr(p) == -1)
+#endif
+			{
+				if (cnf.debug & DEBUG_DNS)
+				log(LOG_DEBUG,
+					"pregetting A of sourceIP: %s", p);
+#ifdef INET6
+				if (cnf.inetdom & SMTP_V4)
+				{
+#endif
+					if (query_domain(p, T_A) < 0)
+						return -1;
+					n++;
+#ifdef INET6
+				}
+				if (cnf.inetdom & SMTP_V6)
+				{
+					if (query_domain(p, T_AAAA) < 0)
+						return -1;
+					n++;
+				}
+#endif
+			}
+			if (q != NULL)
+			{
+				*q = ',';
+				p = q + 1;
+			}
+			else
+				p = NULL;
+		}
+	}
+
 	for (dp = dl; dp != NULL; dp = dp->next)
 	{
 		/* known by previous transaction */
 		if (dp->firstmx != NULL || dp->stat != EX_OK)
 			continue;
 
+#if 0
 		/* skip numeric address spec */
 		if (*(dp->name) == '[')
 			continue;
+#endif
 
 		map_arg = dp->map_arg;
 		while (*map_arg != NULL)
 		{
-			if (strcasecmp(*map_arg, "MX") == 0)
+			int len;
+
+			if ((strcasecmp(*map_arg, "MX") == 0)
+			 || (strcasecmp(*map_arg, "MX?") == 0))
 			{
 				if (cnf.debug & DEBUG_DNS)
 				log(LOG_DEBUG, "pregetting MX: %s", dp->name);
@@ -170,6 +235,58 @@ struct domain *dl;
 				}
 #endif
 			}
+			else if (**map_arg != '[')
+			{
+				if (cnf.debug & DEBUG_DNS)
+				log(LOG_DEBUG, "pregetting MX: %s", *map_arg);
+				if (query_domain(*map_arg, T_MX) < 0)
+					return -1;
+				n++;
+			}
+			else if ((len = strlen(*map_arg)) > 2
+			      && (*map_arg)[len - 1] == ']')
+			{	/* [...] spec */
+				(*map_arg)[len - 1] = '\0';
+#ifdef INET6
+				if (inet_addr((*map_arg) + 1) != -1
+				 || inet_pton(INET6, (*map_arg) + 1,
+					      &dsin.in6.sin6_addr) > 0)
+#else
+				if (inet_addr((*map_arg) + 1) != -1)
+#endif
+					goto skip_preget_a;
+
+				if (cnf.debug & DEBUG_DNS)
+				log(LOG_DEBUG,
+					"pregetting A of sourceIP: %s",
+					(*map_arg) + 1);
+#ifdef INET6
+				if (cnf.inetdom & SMTP_V4)
+				{
+#endif
+					if (query_domain((*map_arg) + 1,
+							 T_A) < 0)
+					{
+						(*map_arg)[len - 1] = ']';
+						return -1;
+					}
+					n++;
+#ifdef INET6
+				}
+				if (cnf.inetdom & SMTP_V6)
+				{
+					if (query_domain((*map_arg) + 1,
+							 T_AAAA) < 0)
+					{
+						(*map_arg)[len - 1] = ']';
+						return -1;
+					}
+					n++;
+				}
+#endif
+  skip_preget_a:
+				(*map_arg)[len - 1] = ']';
+			}
 			map_arg++;
 		}
 	}
@@ -196,7 +313,7 @@ int type;
 
 	qrp = hash_query_lookup(name, &hashp);
 	if (qrp == NULL) {
-		qrp = (struct dns_stat *)malloc(sizeof(struct dns_stat));
+		qrp = (struct dns_stat *)MALLOC(sizeof(struct dns_stat));
 		bzero(qrp, sizeof(struct dns_stat));
 		if (qrp == NULL)
 		{
@@ -298,7 +415,7 @@ int type;
 		return -1;
 	}
 
-	qchain = (struct query_chain *)malloc(sizeof(struct query_chain));
+	qchain = (struct query_chain *)MALLOC(sizeof(struct query_chain));
 	if (qchain == NULL)
 	{
 		/* error */
@@ -308,7 +425,7 @@ int type;
 	bzero(qchain, sizeof(struct query_chain));
 	qchain->chain = qrp->requests;
 	qrp->requests = qchain;
-	qchain->request = (char *)malloc(n);
+	qchain->request = (char *)MALLOC(n);
 	if (qchain->request == NULL)
 	{
 		/* error */
@@ -365,7 +482,7 @@ const struct sockaddr_in *inp;
 	for (ns = 0; ns < _res.nscount; ns++)
 	{
 #if INET6
-		const SockAddr *srv = (SockAddr *)&_res.NSADDR_LIST(ns);
+		const SockAddr *srv = (SockAddr *)&NSADDR_LIST(ns);
 
 		if (srv->in.sin_family == inp->in.sin_family
 		 && srv->in.sin_port == inp->in.sin_port
@@ -380,7 +497,7 @@ const struct sockaddr_in *inp;
 			return ns;
 		}
 #else
-		register const struct sockaddr_in *srv = &_res.NSADDR_LIST(ns);
+		register const struct sockaddr_in *srv = &NSADDR_LIST(ns);
 
 		if (srv->sin_family == inp->sin_family
 		 && srv->sin_port == inp->sin_port
@@ -435,7 +552,7 @@ int sync;
 #if INET6
 	for (n = 0; n < _res.nscount; n++)
 	{
-		const SockAddr *srv = (SockAddr *)&_res.NSADDR_LIST(n);
+		const SockAddr *srv = (SockAddr *)&NSADDR_LIST(n);
 
 		if (s == -1 && srv->in.sin_family == AF_INET)
 		{
@@ -526,7 +643,7 @@ again:
 				itvl = 1;
 			qcp->time = now + itvl;
 #if INET6
-			nsap = (SockAddr *)&_res.NSADDR_LIST(qcp->ns);
+			nsap = (SockAddr *)&NSADDR_LIST(qcp->ns);
 			if (nsap->in.sin_family == AF_INET)
 			{
 				if (cnf.debug & DEBUG_DNS)
@@ -567,7 +684,7 @@ again:
 				}
 			}
 #else
-			nsap = &_res.NSADDR_LIST(qcp->ns);
+			nsap = &NSADDR_LIST(qcp->ns);
 			if (cnf.debug & DEBUG_DNS)
 			log(LOG_DEBUG, "send query n=%s t=%d (s=%s)",
 				qcp->stat->name, qcp->type,
@@ -992,7 +1109,7 @@ int anssiz;
 		ap += n;
 		GETSHORT(type, ap);
 		GETSHORT(class, ap);
-		ap += sizeof(long);
+		ap += INT32SZ;
 		GETSHORT(n, ap);
 		if (type == T_MX)
 		{
@@ -1027,7 +1144,7 @@ int anssiz;
 					nbuf, nbuf2, pref);
 
 				qrp->rr_got |= RR_MX;
-				mxp = (struct mx*)malloc(sizeof(struct mx));
+				mxp = (struct mx*)MALLOC(sizeof(struct mx));
 				mxp->next = qrp->mx;
 				qrp->mx = mxp;
 				mxp->name = newstr(nbuf2);
@@ -1157,7 +1274,7 @@ int anssiz;
 			qrp2 = hash_query_lookup(nbuf2, &hashp);
 			if (qrp2 == NULL) {
 				qrp2 = (struct dns_stat *)
-					malloc(sizeof(struct dns_stat));
+					MALLOC(sizeof(struct dns_stat));
 				bzero(qrp2, sizeof(struct dns_stat));
 				if (qrp2 == NULL)
 				{
@@ -1303,7 +1420,7 @@ int anssiz;
 		qrp = hash_query_lookup(nbuf, &hashp);
 		if (qrp == NULL) {
 			qrp = (struct dns_stat *)
-				malloc(sizeof(struct dns_stat));
+				MALLOC(sizeof(struct dns_stat));
 			bzero(qrp, sizeof(struct dns_stat));
 			if (qrp == NULL)
 			{
@@ -1335,7 +1452,7 @@ int anssiz;
 			qrp = hash_query_lookup(mxp->name, &hashp);
 			if (qrp == NULL) {
 				qrp = (struct dns_stat *)
-					malloc(sizeof(struct dns_stat));
+					MALLOC(sizeof(struct dns_stat));
 				bzero(qrp, sizeof(struct dns_stat));
 				if (qrp == NULL)
 				{
@@ -1400,11 +1517,13 @@ struct domain *dl;
 		while (*map_arg != NULL)
 		{
 			if (strcasecmp(*map_arg, "MX") == 0)
-				rc = getmxrr(dp->name, dp, 1, offset);
+				rc = getmxrr(dp->name, dp, 1, offset, 1);
+			else if (strcasecmp(*map_arg, "MX?") == 0)
+				rc = getmxrr(dp->name, dp, 1, offset, 0);
 			else if (strcasecmp(*map_arg, "A") == 0)
-				rc = getmxrr(dp->name, dp, 0, offset);
+				rc = getmxrr(dp->name, dp, 0, offset, 0);
 			else
-				rc = getmxrr(*map_arg, dp, 1, offset);
+				rc = getmxrr(*map_arg, dp, 1, offset, 0);
 			if (rc < 0 || dp->stat != EX_OK)
 				goto next_domain;
 			if (rc > 0)
@@ -1537,7 +1656,7 @@ struct domain *dp;
 	if (mxp == NULL)
 	{
 		/* new MX */
-		mxp = (struct mx*)malloc(sizeof(struct mx));
+		mxp = (struct mx*)MALLOC(sizeof(struct mx));
 		if (mxp == NULL)
 		{
 			log(LOG_NOTICE, "out of memory (newmx)");
@@ -1561,11 +1680,12 @@ struct domain *dp;
 }
 
 int
-getmxrr(name, dp, query_mx, offset)
+getmxrr(name, dp, query_mx, offset, wait)
 char *name;
 struct domain *dp;
 int query_mx;
 int offset;
+int wait;	/* wait till next trial if MX lookup failure */
 {
 	int seenlocal, localpref, seenfallback;
 	struct mx *mxp, *pmxp, *cmxp;
@@ -1589,23 +1709,52 @@ int offset;
 	{
 		/* numeric address or A RR lookup */
 		int len;
+		SockAddr nsin;
 
-		dp->stat = EX_OK;
 		len = strlen(name);
-		if (len > 2 && name[len-1] == ']' && !isdigit(name[len-2]))
+		if (len <= 2 || name[len-1] != ']')
 		{
-			/* A RR lookup */
-			name[len-1] = '\0';
-			mxp = newmx(name + 1, offset - 1, -1, dp);
-			name[len-1] = ']';
+			dp->stat = EX_NOHOST;
+			snprintf(msgbuf, sizeof(msgbuf),
+				"Bad host name: %s", name);
+			dp->response = newstr(msgbuf);
+			if (dp->response == NULL)
+				return -1;
+			return 0;
 		}
+		dp->stat = EX_OK;
+		name[len-1] = '\0';
+		if ((nsin.in.sin_addr.s_addr = inet_addr(name + 1)) != -1)
+		{
+			/* numeric IPv4 address */
+			name[len-1] = ']';
+			mxp = newmx(name, offset - 1, -1, dp);
+			if (mxp == NULL)
+				return -1;
+			addinetaddress(mxp, AF_INET, sizeof(nsin.in.sin_addr),
+				&nsin.in.sin_addr);
+		}
+#ifdef INET6
+		else if (inet_pton(INET6, name + 1, &nsin.in6.sin6_addr) > 0)
+		{
+			/* numeric IPv6 address */
+			name[len-1] = ']';
+			mxp = newmx(name, offset - 1, -1, dp);
+			if (mxp == NULL)
+				return -1;
+			addinetaddress(mxp, AF_INET6,
+				sizeof(nsin.in6.sin6_addr),
+				&nsin.in6.sin6_addr);
+		}
+#endif
 		else
 		{
-			/* numeric address */
-			mxp = newmx(name, offset - 1, -1, dp);
+			/* A RR lookup */
+			mxp = newmx(name + 1, offset - 1, -1, dp);
+			name[len-1] = ']';
+			if (mxp == NULL)
+				return -1;
 		}
-		if (mxp == NULL)
-			return -1;
 		return 1;	/* find adderess */
 	}
 
@@ -1636,13 +1785,18 @@ int offset;
 
 	if (qrp == NULL || qrp->rr_timeout & RR_MX) {
 		/* couldn't connect to the name server */
-		if (*fallbackmx)
+		if (cnf.fallbackmx != NULL)
 		{
 			dp->stat = EX_OK;
-			mxp = newmx(fallbackmx, offset - 1, 0, dp);
+			mxp = newmx(cnf.fallbackmx, offset - 1, 0, dp);
 			if (mxp == NULL)
 				return -1;
-			return 1;	/* find adderess */
+			return 1;	/* adderess found */
+		}
+		if (!wait)
+		{
+			dp->stat = EX_OK;
+			return 1;
 		}
 		dp->stat = EX_TEMPFAIL;
 		dp->response = "DNS temporary lookup failure";
@@ -1706,7 +1860,8 @@ int offset;
 			seenlocal = 1;
 			continue;
 		}
-		if (strcasecmp(mxname, fallbackmx) == 0)
+		if (cnf.fallbackmx != NULL
+		 && strcasecmp(mxname, cnf.fallbackmx) == 0)
 			seenfallback = 1;
 
 		/* truncate lower prefered MX (imperfect) */
@@ -1760,7 +1915,8 @@ int offset;
 			continue;
 		}
 
-		mxp = newmx(mxname, offset + cmxp->pref, mxrand(mxname), dp);
+		mxp = newmx(mxname, offset + cmxp->pref,
+			mxrand(mxname, qrp->name), dp);
 		if (mxp == NULL)
 			return -1;
 	}
@@ -1806,9 +1962,9 @@ int offset;
 		}
 	}
 
-	if (*fallbackmx && !seenfallback)
+	if (cnf.fallbackmx != NULL && !seenfallback)
 	{
-		mxp = newmx(fallbackmx, offset + 256, 0, dp);
+		mxp = newmx(cnf.fallbackmx, offset + 256, 0, dp);
 		if (mxp == NULL)
 			return -1;
 	}
@@ -1865,7 +2021,7 @@ u_char *t, *b, *e;
 		ap += n;
 		GETSHORT(class, ap);
 		GETSHORT(type, ap);
-		ap += sizeof(long);
+		ap += INT32SZ;
 		GETSHORT(n, ap);
 		if (type == T_MX) {
 			GETSHORT(pref, ap);
