@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: smtp.c,v 1.117 1998/09/30 14:04:27 motonori Exp $";
+static char *_id_ = "$Id: smtp.c,v 1.130 1999/04/15 05:43:17 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -47,7 +47,7 @@ static	int num_trans, num_sock, max_sock;
 static	jmp_buf jmp_env;
 static	time_t Now;
 static	int FirstRead, FirstWrite;
-static	int Terminate, NewTransCheck, NeedTrimDomain;
+static	int Terminate, NewTransCheck, NeedTrimDomain, MoreFreeSlot;
 
 static void
 sig_alarm()
@@ -94,6 +94,50 @@ deliver()
 		env.sender, env.realsize, sti.nrcpt,
 		(env.mid == NULL)?"none":env.mid);
 
+	if (cnf.showrecipient > 0 && cnf.src_mod_path != NULL)
+	{
+		FILE *fp;
+
+		if (cnf.debug & DEBUG_TRANS)
+			log(LOG_DEBUG, "searching sender address %s in %s",
+				env.sender, cnf.src_mod_path);
+		if ((fp = fopen(cnf.src_mod_path, "r")) == NULL)
+		{
+			/* file missing */
+#if 0
+			cnf.showrecipient = 0;
+#endif
+		}
+		else
+		{
+			char buf[MAXLINE];
+			int found = 0;
+			int ll;
+
+			while (fgets(buf, sizeof(buf), fp) != NULL)
+			{
+				if ((ll = strlen(buf)) > 0 && buf[ll-1] == '\n')
+					buf[ll-1] = '\0';
+				if (strstr(env.sender, buf) != NULL)
+				{
+					found = 1;
+					break;
+				}
+			}
+			fclose(fp);
+			if (found)
+			{
+				log(LOG_INFO,
+					"recipients/transaction changed to 1");
+				cnf.rcpts_trans == 1;
+			}
+#if 0
+			else
+				cnf.showrecipient = 0;
+#endif
+		}
+	}
+
 	bzero(connwork, sizeof(connwork));
 
 	sti.time_start = time(NULL); 
@@ -132,6 +176,7 @@ deliver()
 	Terminate = 0;
 	NewTransCheck = 1;
 	NeedTrimDomain = 0;
+	MoreFreeSlot = 0;
 
 	/* process delivery with SMTP */
 	sti.ntrans = 0;
@@ -289,10 +334,9 @@ smtp()
 	struct domain *domp, *wdomp;
 	struct timeval timeout;
 	fd_set r_fds, w_fds;
-	int i, fds, rc, in_use;
+	int i, fds, rc, in_use, burst, true_terminate;
 	int all_done = 0;
 	int bad_state = 0;
-	int true_terminate;
 	char *msg;
 
 	/* initialize work pointers */
@@ -422,16 +466,22 @@ smtp()
 		}
 
 		/* create new transaction if there is a free port */
+		burst = 0;
+		MoreFreeSlot = 0;
 		while (NewTransCheck
 		    && num_trans <= (max_sock - sti.minsock)
 		    && num_sock <= (max_sock - sti.minsock))
 		{
+			if (++burst > 16)	/* slow start */
+			{
+				MoreFreeSlot = 1;
+				break;
+			}
 #if 1
 			if (cnf.debug & DEBUG_TRANS)
 			log(LOG_DEBUG, "new transaction: trans=%d, sock=%d",
 				num_trans, num_sock);
 #endif
-
 			if (new_transaction() < 0)
 				break;
 		}
@@ -488,7 +538,10 @@ smtp()
 		else
 		{
 			timeout.tv_usec = 0;
-			timeout.tv_sec = 10;
+			if (MoreFreeSlot)
+				timeout.tv_sec = 1;
+			else
+				timeout.tv_sec = 10;
 			rc = select(fds+1, &r_fds, &w_fds, NULL, &timeout);
 		}
 
@@ -999,6 +1052,7 @@ int tracechain;
 		else
 			setproctitle("%d%% delivered", n);
 	}
+	lmtp_data_response(0);
 }
 
 static struct connection *
@@ -1498,6 +1552,12 @@ struct connection *conn;
 			max_sock, num_sock, sti.minsock);
 		if (max_sock > NOFILE)
 			max_sock = sti.maxsock;
+#if 1
+		if ((errno == ENFILE || errno == EMFILE)
+		 && max_sock > sti.maxsock && sti.maxsock - 10 > sti.minsock)
+			/* reduce max socket number */
+			max_sock = sti.maxsock - 10;
+#endif
 		/* XXX give up all new transaction */
 		/* no markstatus */
 		conn->host->state = STAT_CLOSED;
@@ -1520,22 +1580,97 @@ struct connection *conn;
 	}
 #endif
 
-	if (sourceIP[0] != '\0')
+	found = 0;
+	if (cnf.sourceIP != NULL)
 	{
 		/* bind source address */
-		switch (conn->host->current->domain)
-		{
-		    case AF_INET:
-			lsin.in.sin_addr.s_addr = inet_addr(sourceIP);
-			break;
+		char *p, *q;
+		struct dns_stat *qrp;
+		struct host *hostp;
+		struct inetaddr *addrp;
 #ifdef INET6
-		    case AF_INET6:
-			inet_pton(INET6, sourceIP, &lsin.in6.sin6_addr);
-			break;
+		char buf[64];
 #endif
+
+		p = cnf.sourceIP;
+		while (p != NULL)
+		{
+			q = strchr(p, ',');
+			if (q != NULL)
+				*q = '\0';
+
+			if ((qrp = hash_query_lookup(p, NULL)) == NULL)
+			{
+				/* no query made */
+
+				switch (conn->host->current->domain)
+				{
+				    case AF_INET:
+					lsin.in.sin_addr.s_addr = inet_addr(p);
+					if (lsin.in.sin_addr.s_addr != -1)
+						found = 1;
+					break;
+#ifdef INET6
+				    case AF_INET6:
+					if (inet_pton(INET6, p,
+						      &lsin.in6.sin6_addr) > 0)
+						found = 1;
+					break;
+#endif
+				}
+				goto sip_next;
+			}
+
+			while (qrp->cname != NULL)
+				qrp = qrp->cname;
+			if ((hostp = hash_host_lookup(qrp->name, NULL)) == NULL)
+				goto sip_next;
+
+			addrp = hostp->addr;
+			for (; addrp != NULL; addrp = addrp->next)
+			{
+				if (conn->host->current->domain != addrp->domain)
+					continue;
+
+				found = 1;
+				switch (addrp->domain)
+				{
+				    case AF_INET:
+					bcopy(addrp->address,
+						&lsin.in.sin_addr, addrp->len);
+					log(LOG_INFO,
+						"using %s as source IP address",
+						inet_ntoa(lsin.in.sin_addr));
+					break;
+#ifdef INET6
+				    case AF_INET6:
+					bcopy(addrp->address,
+						&lsin.in6.sin6_addr,
+						addrp->len);
+
+					inet_ntop(AF_INET6, &lsin.in6.sin6_addr,
+						buf, sizeof(buf));
+					log(LOG_INFO, "using IPv6:%s as source IP address", buf);
+					break;
+#endif
+				}
+				break;
+			}
+  sip_next:
+			if (found)
+				break;
+			if (q != NULL)
+			{
+				*q = ',';
+				p = q + 1;
+			}
+			else
+				p = NULL;
 		}
+		if (!found)
+			log(LOG_INFO, "sourceIP (%s) ignored", cnf.sourceIP);
 	}
-	else
+	if (!found)
 	{
 		switch (conn->host->current->domain)
 		{
@@ -1544,8 +1679,7 @@ struct connection *conn;
 			break;
 #ifdef INET6
 		    case AF_INET6:
-			lsin.in6.sin6_addr = in6addr_any;
-			break;
+			lsin.in6.sin6_addr = in6addr_any; break;
 #endif
 		}
 	}
@@ -1579,7 +1713,7 @@ struct connection *conn;
 		}
 	}
 
-	if (sourceIP[0] != '\0' || cnf.src_port > 0)
+	if (cnf.sourceIP != NULL || cnf.src_port > 0)
 	{
 		if (bind(sd, (struct sockaddr *)&lsin, socksize) < 0)
 		{
@@ -1738,10 +1872,10 @@ struct connection *conn;
 	/* create buffer */
 	conn->bufsize = SMTPBUFSIZE;
 	if (conn->inbuf == NULL)
-		conn->inbuf = (char *)malloc(conn->bufsize);
+		conn->inbuf = (char *)MALLOC(conn->bufsize);
 	conn->inlen = 0;
 	if (conn->inbuf != NULL && conn->outbuf == NULL)
-		conn->outbuf = (char *)malloc(conn->bufsize);
+		conn->outbuf = (char *)MALLOC(conn->bufsize);
 	conn->outlen = 0;
 	if (conn->inbuf == NULL || conn->outbuf == NULL)
 	{
@@ -2105,7 +2239,7 @@ int flush;
 	else
 	{
 		char *p;
-		p = (char *)malloc(len + 3);
+		p = (char *)MALLOC(len + 3);
 		if (p == NULL)
 		{
 			forced_terminate();
@@ -2457,46 +2591,107 @@ struct connection *conn;
 {
 	char cmdbuf[MAXLINE];
 	int len = sizeof(cmdbuf) - 1;
+	int ll;
 
 	strcpy(cmdbuf, "MAIL FROM:");
 	len -= 10;
-	if (len < strlen(env.sender))
-		goto mail_err;
-	strcat(cmdbuf, env.sender);
-	len -= strlen(env.sender);
+
+	if (cnf.showrecipient == 2 && cnf.rcpts_trans == 1
+	 && conn->rcpt->chain == NULL)
+	{
+		/* only one recipient */
+		char su[MAXLINE], sd[MAXLINE];
+		char ru[MAXLINE], rd[MAXLINE];
+		char na[MAXLINE*2];
+		char *p, *q;
+
+		if (strlen(env.sender) > MAXLINE)
+			goto mail_err;
+		p = env.sender;
+		while (*p == '<')
+			p++;
+		q = su;
+		while (*p != '\0' && *p != '@' && *p != '>')
+			*q++ = *p++;
+		*q = '\0';
+		if (*p++ == '@')
+		{
+			q = sd;
+			while (*p != '\0' && *p != '>')
+				*q++ = *p++;
+			*q = '\0';
+		}
+		else
+		{
+			*sd = '\0';
+		}
+		if (strlen(conn->crcpts->address) > MAXLINE)
+			goto mail_err;
+		p = conn->crcpts->address;
+		while (*p == '<')
+			p++;
+		q = ru;
+		while (*p != '\0' && *p != '@' && *p != '>')
+			*q++ = *p++;
+		*q = '\0';
+		if (*p++ == '@')
+		{
+			q = rd;
+			while (*p != '\0' && *p != '>')
+				*q++ = *p++;
+			*q = '\0';
+		}
+		else
+		{
+			*rd = '\0';
+		}
+
+		snprintf(na, sizeof(na), cnf.src_mod_ptn, su, ru, rd, sd);
+		if (len < (ll = strlen(na)))
+			goto mail_err;
+		strcat(cmdbuf, na);
+		len -= ll;
+	}
+	else
+	{
+		if (len < (ll = strlen(env.sender)))
+			goto mail_err;
+		strcat(cmdbuf, env.sender);
+		len -= ll;
+	}
 
 	if (env.envid != NULL && conn->host->proto & PROTO_DSN)
 	{
-		if (len < strlen(env.envid) + 7)
+		if (len < (ll = strlen(env.envid) + 7))
 			goto mail_err;
 		strcat(cmdbuf, " ENVID=");
 		strcat(cmdbuf, env.envid);
-		len -= strlen(env.envid) + 7;
+		len -= ll;
 	}
 	if (env.size != NULL && conn->host->proto & PROTO_SIZE)
 	{
-		if (len < strlen(env.size) + 6)
+		if (len < (ll = strlen(env.size) + 6))
 			goto mail_err;
 		/* XXX should send REAL message size */
 		strcat(cmdbuf, " SIZE=");
 		strcat(cmdbuf, env.size);
-		len -= strlen(env.size) + 6;
+		len -= ll;
 	}
 	if (env.body != NULL && conn->host->proto & PROTO_8BITMIME)
 	{
-		if (len < strlen(env.body) + 6)
+		if (len < (ll = strlen(env.body) + 6))
 			goto mail_err;
 		strcat(cmdbuf, " BODY=");
 		strcat(cmdbuf, env.body);
-		len -= strlen(env.body) + 6;
+		len -= ll;
 	}
 	if (env.ret != NULL && conn->host->proto & PROTO_DSN)
 	{
-		if (len < strlen(env.ret) + 5)
+		if (len < (ll = strlen(env.ret) + 5))
 			goto mail_err;
 		strcat(cmdbuf, " RET=");
 		strcat(cmdbuf, env.ret);
-		len -= strlen(env.ret) + 5;
+		len -= ll;
 	}
 
 	if (smtp_send_command(conn, cmdbuf, sizeof(cmdbuf),
@@ -2611,7 +2806,7 @@ smtp_rcpt_s(conn)
 struct connection *conn;
 {
 	char cmdbuf[MAXLINE];
-	int rcode, len;
+	int rcode, len, ll;
 
 	if (conn->crcpts == NULL)
 	{
@@ -2624,26 +2819,26 @@ struct connection *conn;
 	len = sizeof(cmdbuf) - 1;
 	strcpy(cmdbuf, "RCPT TO:");
 	len -= 8;
-	if (len < strlen(conn->crcpts->address))
+	if (len < (ll = strlen(conn->crcpts->address)))
 		goto rcpt_err;
 	strcat(cmdbuf, conn->crcpts->address);
-	len -= strlen(conn->crcpts->address);
+	len -= ll;
 
 	if (conn->crcpts->notify != NULL && conn->host->proto & PROTO_DSN)
 	{
-		if (len < strlen(conn->crcpts->notify) + 8)
+		if (len < (ll = strlen(conn->crcpts->notify) + 8))
 			goto rcpt_err;
 		strcat(cmdbuf, " NOTIFY=");
 		strcat(cmdbuf, conn->crcpts->notify);
-		len -= strlen(conn->crcpts->notify) + 8;
+		len -= ll;
 	}
 	if (conn->crcpts->orcpt != NULL && conn->host->proto & PROTO_DSN)
 	{
-		if (len < strlen(conn->crcpts->orcpt) + 7)
+		if (len < (ll = strlen(conn->crcpts->orcpt) + 7))
 			goto rcpt_err;
 		strcat(cmdbuf, " ORCPT=");
 		strcat(cmdbuf, conn->crcpts->orcpt);
-		len -= strlen(conn->crcpts->orcpt) + 7;
+		len -= ll;
 	}
 
 	if (conn->host->proto & PROTO_PIPELINING)
@@ -2906,7 +3101,8 @@ struct connection *conn;
 	char *datap, *p, savec;
 	int rcode, len;
 
-	if (cnf.showrecipient && conn->rcpt->chain == NULL /* only one rcpt. */
+	if (cnf.showrecipient == 1 && cnf.rcpts_trans == 1
+	 && conn->rcpt->chain == NULL /* only one recipient */
 	 && conn->msg->data == env.msg->data && conn->offset == 0)
 	{
 		char altbuf[MAXLINE];
